package ru.dbotthepony.kstarbound.defs.image

import com.google.gson.Gson
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.io.json.consumeNull
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.SBPattern

/**
 * @see [AtlasConfiguration.Companion.get]
 */
class ImageReference private constructor(
	val raw: AssetPath,
	val imagePath: SBPattern,
	val spritePath: SBPattern?,
	val atlas: AtlasConfiguration?,
) {
	/**
	 * Спрайт, на которое ссылается данный референс, или `null` если:
	 * * [atlas] равен `null`
	 * * [spritePath] является шаблоном и определены не все значения
	 * * [spritePath] не является правильным именем спрайта внутри [atlas] (смотрим [AtlasConfiguration.get])
	 */
	val sprite by lazy {
		if (atlas == null)
			null
		else if (spritePath == null)
			atlas.any()
		else
			atlas[spritePath.value ?: return@lazy null]
	}

	fun with(values: (String) -> String?): ImageReference {
		val imagePath = this.imagePath.with(values)
		var spritePath = this.spritePath?.with(values)

		if (spritePath == null) {
			val frame = values.invoke("frame")

			if (frame != null) {
				spritePath = SBPattern.FRAME.with(values)
			}
		}

		if (imagePath != this.imagePath || spritePath != this.spritePath) {
			if (imagePath != this.imagePath) {
				val resolved = imagePath.value

				if (resolved == null)
					return ImageReference(raw, imagePath, spritePath, null)
				else
					return ImageReference(raw, imagePath, spritePath, AtlasConfiguration.get(resolved))
			} else {
				return ImageReference(raw, imagePath, spritePath, atlas)
			}
		}

		return this
	}

	fun with(values: Map<String, String>): ImageReference {
		return with(values::get)
	}

	override fun equals(other: Any?): Boolean {
		return other is ImageReference && other.imagePath == imagePath && other.spritePath == spritePath && other.atlas == atlas
	}

	override fun hashCode(): Int {
		return imagePath.hashCode() * 31 + spritePath.hashCode()
	}

	override fun toString(): String {
		return "ImageReference[$imagePath:$spritePath]"
	}

	companion object : TypeAdapter<ImageReference>() {
		val NEVER = ImageReference(AssetPath("", ""), SBPattern.EMPTY, null, null)

		private val strings by lazy { Starbound.gson.getAdapter(String::class.java) }

		override fun write(out: JsonWriter, value: ImageReference?) {
			out.value(value?.raw?.fullPath)
		}

		@JvmStatic
		fun create(path: String): ImageReference {
			if (path == "")
				return NEVER

			val split = path.split(':')

			if (split.size > 2) {
				throw JsonSyntaxException("Ambiguous image reference: $path")
			}

			val imagePath = if (split.size == 2) SBPattern.of(split[0]) else SBPattern.of(path)
			val spritePath = if (split.size == 2) SBPattern.of(split[1]) else null

			if (imagePath.isPlainString) {
				val remapped = AssetPathStack.remap(split[0])
				return ImageReference(AssetPath(path, AssetPathStack.remap(path)), SBPattern.raw(remapped), spritePath, AtlasConfiguration.get(remapped))
			} else {
				return ImageReference(AssetPath(path, path), imagePath, spritePath, null)
			}
		}

		override fun read(`in`: JsonReader): ImageReference? {
			if (`in`.consumeNull())
				return null

			return create(strings.read(`in`))
		}
	}
}
